#define APP_NAME "/apps/compiz"
/*
 * Copyright © 2006 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* This is a hack to generate CSM schema files for compiz and its
 * plugins. To regenerate the base settings.in file, just do
 *
 *   rm settings.in; make settings.in
 *
 * To generate a schema file for third-party plugins, do something
 * like:
 *
 *   COMPIZ_SCHEMA_PLUGINS="plugin1 plugin2" \
 *   COMPIZ_SCHEMA_FILE="output_filename" \
 *   compiz --replace dep1 plugin1 dep2 dep3 plugin2 csm-dump
 *
 * COMPIZ_SCHEMA_PLUGINS indicates the plugins to generate schema
 * info for, which might be a subset of the plugins listed on the
 * compiz command line. (Eg, if your plugin depends on "cube", you
 * need to list that on the command line, but don't put in the
 * environment variable, because you don't want to dump the cube schema
 * into your plugin's schema file.)
 */

#include <config.h>

#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <math.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <compiz.h>
#undef _
#undef N_

#include <glib.h>
#include <glib/gi18n.h>

#include "csm-utils.h"

static GKeyFile *schemaFile;

static int displayPrivateIndex;

typedef struct _CSMDumpDisplay {
    HandleEventProc handleEvent;
} CSMDumpDisplay;

#define GET_GCONF_DUMP_DISPLAY(d)				  \
    ((CSMDumpDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define GCONF_DUMP_DISPLAY(d)			      \
    CSMDumpDisplay *gd = GET_GCONF_DUMP_DISPLAY (d)

/* Decimal point character */
#define DECIMAL_POINT _(".")

static char **locales;

static char *
csmTypeToString (CompOptionType type)
{
    switch (type) {
    case CompOptionTypeBool:
	return "bool";
    case CompOptionTypeColor:
	return "color";
    case CompOptionTypeInt:
	return "int";
    case CompOptionTypeFloat:
	return "float";
    case CompOptionTypeString:
	return "string";
    case CompOptionTypeList:
	return "list";
    case CompOptionTypeAction:
    return "action";
    default:
	break;
    }

    return "unknown";
}

#define COMP_OPTION_IS_NUMERIC(o) \
	(isdigit ((unsigned char)o->name[strlen (o->name) - 1]))

static char *
translate_numeric (const char *str)
{
    char *p, *q, *denum, *out;

    p = strpbrk (str, "0123456789");
    if (!p)
	return g_strdup (str);
    q = p + strspn (p, "0123456789");

    denum = g_strdup_printf ("%.*s%%d%s", (int)(p - str), str, q);
    if (strcmp (denum, _(denum)) != 0)
	out = g_strdup_printf (_(denum), strtol (p, NULL, 10));
    else
	out = g_strdup (str);
    g_free (denum);

    return out;
}

static char *
getShortDesc (CompOption *o)
{
    if (COMP_OPTION_IS_NUMERIC (o))
	return translate_numeric (o->shortDesc);
    else
	return g_strdup (_(o->shortDesc));
}

static char *
getLongDesc (CompOption *o, const char *option_name)
{
    char *base;

    if (COMP_OPTION_IS_NUMERIC (o))
	base = translate_numeric (o->longDesc);
    else
	base = g_strdup (_(o->longDesc));
    return base;
}

static void
csmDumpToSchema (CompDisplay *d,
		   CompOption  *o,
		   char	       *name,
		   char	       *plugin,
		   char	       *screen)
{
    gchar * sect;
    gchar * basename;
    gchar *cshort, *clong;

    if (plugin && screen)
    {
        sect=plugin;
        if (strcmp(screen,"allscreens")==0)
            basename=g_strdup_printf("a_%s",name);
        else
            basename=g_strdup_printf("s_%s",name);
    }
    else if (plugin)
    {
        sect=plugin;
        basename=g_strdup_printf("p_%s",name);
    }
    else
    {
        sect="_";
        if (strcmp(screen,"allscreens")==0)
            basename=g_strdup_printf("a_%s",name);
        else
            basename=g_strdup_printf("s_%s",name);
    }
    
    
    g_key_file_set_string(schemaFile,sect,
            g_strconcat("o_",basename,NULL),csmTypeToString(o->type));
    if (o->type == CompOptionTypeList)
    {
        g_key_file_set_string(schemaFile,sect,g_strconcat("t_",basename,NULL),
                csmTypeToString(o->value.list.type));
    }
    if (o->type == CompOptionTypeAction)
    {
        gboolean states[] = {FALSE,FALSE,FALSE,FALSE};
        //check for action types here
        if (o->value.action.state & CompActionStateInitButton)
            states[0]=TRUE;
        if (o->value.action.state & CompActionStateInitKey)
            states[1]=TRUE;
        if (o->value.action.state & CompActionStateInitEdge)
            states[2]=TRUE;
        if (o->value.action.state & CompActionStateInitBell)
            states[3]=TRUE;
        g_key_file_set_boolean_list(schemaFile,sect,
                g_strconcat("s_",basename,NULL),states,4);
    }
    else if (o->type == CompOptionTypeString || (o->type == CompOptionTypeList && 
                o->value.list.type == CompOptionTypeString))
    {
        if (o->rest.s.nString)
        {
            g_key_file_set_string_list(schemaFile,sect,
                    g_strconcat("s_",basename,NULL),
                    (const gchar **) o->rest.s.string,o->rest.s.nString);
        }
    }
    else if (o->type == CompOptionTypeInt)
    {
        gint n[]={o->rest.i.min,o->rest.i.max};
        g_key_file_set_integer_list(schemaFile,sect,
                g_strconcat("s_",basename,NULL),
                n,2);
    }
    else if (o->type == CompOptionTypeFloat)
    {
        gchar * s[3];
        gint i;
        for (i=0;i<3;i++)
        {
            s[i]=malloc(G_ASCII_DTOSTR_BUF_SIZE);
        }
        g_ascii_dtostr(s[0],G_ASCII_DTOSTR_BUF_SIZE,o->rest.f.min);
        g_ascii_dtostr(s[1],G_ASCII_DTOSTR_BUF_SIZE,o->rest.f.max);
        g_ascii_dtostr(s[2],G_ASCII_DTOSTR_BUF_SIZE,o->rest.f.precision);
        g_key_file_set_string_list(schemaFile,sect,
                g_strconcat("s_",basename,NULL),
                (const gchar **) s,3);
    }
    gchar * dval=g_strconcat("d_",basename,NULL);
    if (strcmp(sect,"_")==0 && strcmp(basename,"a_active_plugins")==0)
        dval=g_strconcat("h_",basename,NULL);
    csm_save_option(schemaFile,sect,dval,o,d);

    setlocale (LC_ALL, "C");
    cshort = getShortDesc (o);
    clong = getLongDesc (o, name);
    g_key_file_set_string(schemaFile,sect,g_strconcat("n_",basename,NULL),cshort);
    g_key_file_set_string(schemaFile,sect,g_strconcat("l_",basename,NULL),clong);
}

static void
dumpGeneralOptions (CompDisplay *d)
{
    CompOption   *option;
    int	         nOption;

    option = compGetDisplayOptions (d, &nOption);
    while (nOption--)
    {
	if (!strcmp (option->name, "active_plugins"))
	{
        static gchar * xPlugins[]={"bench","blur","showdesktop","plane","animation"}; // remove these from d_
#define nXPlug (sizeof(xPlugins)/sizeof(gchar *))
        gchar ** iplugins = malloc(sizeof(gchar *)*option->value.list.nValue);
        gint nP=0;
        gint i;
	    /* "Fix" it for the schema file */
	    memmove (option->value.list.value + 2,
		     option->value.list.value+1,
		     (option->value.list.nValue - 2) *
		     sizeof (CompOptionValue));
	    option->value.list.value[1].s = strdup ("csm");
        for (i=0;i<option->value.list.nValue;i++)
        {
            gboolean inc=TRUE;
            gint j;
            for (j=0;j<nXPlug;j++)
                if (strcmp(xPlugins[j],option->value.list.value[i].s)==0)
                    inc=FALSE;
            if (inc)
                iplugins[nP++]=option->value.list.value[i].s;
        }
        g_key_file_set_string_list(schemaFile,"_","d_a_active_plugins",(const gchar **) iplugins,nP);
	}

    csmDumpToSchema (d, option, option->name, NULL, "allscreens");

	option++;
    }

    option = compGetScreenOptions (&d->screens[0], &nOption);
    while (nOption--)
    {
	    csmDumpToSchema (d, option, option->name, NULL, "screen0");
	option++;
    }
}

static void
dumpPluginOptions (CompDisplay *d, CompPlugin *p)
{
    CompOption   *option, info, bopt, ropt;
    int	         nOption;
    GArray       *reqs, *before;

    if (!strcmp (p->vTable->name, "csm-dump"))
	return;

    memset (&info, 0, sizeof (info));
    info.name      = "info";
    info.shortDesc = p->vTable->shortDesc;
    info.longDesc  = p->vTable->longDesc;
    info.type      = CompOptionTypeString;
    info.value.s   = "";
    csmDumpToSchema (d, &info, info.name, p->vTable->name, NULL);

    reqs   = g_array_new (FALSE, FALSE, sizeof (CompOptionValue));
    before = g_array_new (FALSE, FALSE, sizeof (CompOptionValue));

    if (p->vTable->deps)
    {
	int             i;
	CompOptionValue value;

	for (i = 0; i < p->vTable->nDeps; i++)
	{
	    value.s = p->vTable->deps[i].plugin;
	    if (p->vTable->deps[i].rule == CompPluginRuleBefore)
		g_array_append_val (before, value);
	    else
		g_array_append_val (reqs, value);
	}
    }

    memset (&bopt, 0, sizeof (bopt));
    bopt.name		   = "load_before";
    bopt.shortDesc	   = _("Plugins that this must load before");
    bopt.longDesc	   = _("Do not modify");
    bopt.type		   = CompOptionTypeList;
    bopt.value.list.type   = CompOptionTypeString;
    bopt.value.list.nValue = before->len;
    bopt.value.list.value  = (CompOptionValue *)before->data;
    csmDumpToSchema (d, &bopt, bopt.name, p->vTable->name, NULL);

    memset (&ropt, 0, sizeof (ropt));
    ropt.name		   = "requires";
    ropt.shortDesc	   = _("Plugins that this requires");
    ropt.longDesc	   = _("Do not modify");
    ropt.type		   = CompOptionTypeList;
    ropt.value.list.type   = CompOptionTypeString;
    ropt.value.list.nValue = reqs->len;
    ropt.value.list.value  = (CompOptionValue *)reqs->data;
    csmDumpToSchema (d, &ropt, ropt.name, p->vTable->name, NULL);

    g_array_free (before, TRUE);
    g_array_free (reqs, TRUE);

    if (p->vTable->getDisplayOptions)
    {
	option = p->vTable->getDisplayOptions (d, &nOption);
	while (nOption--)
	{
		csmDumpToSchema (d, option, option->name, p->vTable->name,
				   "allscreens");

	    option++;
	}
    }

    if (p->vTable->getScreenOptions)
    {
	option = p->vTable->getScreenOptions (&d->screens[0], &nOption);
	while (nOption--)
	{
		csmDumpToSchema (d, option, option->name, p->vTable->name,
				   "screen0");

	    option++;
	}
    }
}

static int
strcmpref (const void *p1, const void *p2)
{
    return strcmp (*(const char **)p1, *(const char **)p2);
}

static int
dumpSchema (CompDisplay *d)
{
    const char *schemaFilename, *pluginList;
    char **plugins;
    int i;

    if (findActivePlugin ("csm"))
    {
	fprintf (stderr, "Can't use csm-dump plugin with csm plugin\n");
	return 1;
    }

    pluginList = getenv ("COMPIZ_SCHEMA_PLUGINS");
    if (!pluginList)
    {
	fprintf (stderr, "COMPIZ_SCHEMA_PLUGINS must be set\n");
	return 1;
    }

    plugins = g_strsplit (pluginList, " ", 0);
    qsort (plugins, g_strv_length (plugins), sizeof (char *), strcmpref);

    schemaFilename = getenv ("COMPIZ_SCHEMA_FILE");
    if (!schemaFilename)
	schemaFilename = "settings.dump";

    schemaFile = g_key_file_new();
        //fopen (schemaFilename, "w");

    if (getenv ("COMPIZ_SCHEMA_GENERAL"))
	dumpGeneralOptions (d);

    for (i = 0; plugins[i]; i++)
    {
	CompPlugin *plugin = findActivePlugin (plugins[i]);
	if (!plugin)
	{
	    fprintf (stderr, "No such plugin %s\n", plugins[i]);
	    return 1;
	}
	dumpPluginOptions (d, plugin);
    }

    if (!g_file_set_contents(schemaFilename,g_key_file_to_data(schemaFile,NULL,NULL),-1,NULL))
    {
	fprintf (stderr, "Could not open %s: %s\n", schemaFilename,
		 strerror (errno));
	return 1;
    }

    return 0;
}

static char *restartArgv[] = { "--replace","dbus", "csm", NULL };

static void
csmDumpHandleEvent (CompDisplay *d,
		      XEvent	  *event)
{
    int status;

    GCONF_DUMP_DISPLAY (d);

    UNWRAP (gd, d, handleEvent);

    status = dumpSchema (d);

    if (fork () != 0)
	exit (status);

    programArgv = restartArgv;
    kill (getpid (), SIGHUP);
}

static Bool
csmDumpInitDisplay (CompPlugin  *p,
		      CompDisplay *d)
{
    CSMDumpDisplay *gd;

    gd = malloc (sizeof (CSMDumpDisplay));
    if (!gd)
	return FALSE;

    WRAP (gd, d, handleEvent, csmDumpHandleEvent);

    d->privates[displayPrivateIndex].ptr = gd;

    return TRUE;
}

static void
csmDumpFiniDisplay (CompPlugin  *p,
		      CompDisplay *d)
{
    GCONF_DUMP_DISPLAY (d);

    UNWRAP (gd, d, handleEvent);
    free (gd);
}

static Bool
csmDumpInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    locales = g_strsplit (ALL_LINGUAS, " ", 0);
    setenv ("LANG", "en_US.UTF-8", TRUE);
    setlocale (LC_ALL, "");
    bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    return TRUE;
}

static void
csmDumpFini (CompPlugin *p)
{
    if (displayPrivateIndex >= 0)
	freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginVTable csmDumpVTable = {
    "csm-dump",
    "CSM dump",
    "CSM dump - dumps csm schemas",
    csmDumpInit,
    csmDumpFini,
    csmDumpInitDisplay,
    csmDumpFiniDisplay,
    0, /* InitScreen */
    0, /* FiniScreen */
    0, /* InitWindow */
    0, /* FiniWindow */
    0, /* GetDisplayOptions */
    0, /* SetDisplayOption */
    0, /* GetScreenOptions */
    0, /* SetScreenOption */
    0,
    0
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &csmDumpVTable;
}
