/*
 * Copyright © 2005 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <davidr@novell.com>
 */

#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <compiz.h>

#include <glib.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>
#include "csm-utils.h"

#define KEY_CHANGE_TIMEOUT 250

static int displayPrivateIndex;
enum _foo
{
    CSM_DISPLAY_OPTION_RELOAD,
    CSM_DISPLAY_OPTION_NUM
};
typedef struct _CSMDisplay {
    int screenPrivateIndex;

    GKeyFile * schemaFile; // defaults in here
    GKeyFile * settingsFile; // user settinge here

    CompTimeoutHandle timeoutHandle;

    InitPluginForDisplayProc      initPluginForDisplay;
    SetDisplayOptionProc	  setDisplayOption;
    SetDisplayOptionForPluginProc setDisplayOptionForPlugin;

    CompOption opt[CSM_DISPLAY_OPTION_NUM];
} CSMDisplay;

typedef struct _CSMScreen {
    InitPluginForScreenProc      initPluginForScreen;
    SetScreenOptionProc		 setScreenOption;
    SetScreenOptionForPluginProc setScreenOptionForPlugin;
} CSMScreen;

#define GET_CSM_DISPLAY(d)				      \
    ((CSMDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define CSM_DISPLAY(d)		     \
    CSMDisplay *gd = GET_CSM_DISPLAY (d)

#define GET_CSM_SCREEN(s, gd)				         \
    ((CSMScreen *) (s)->privates[(gd)->screenPrivateIndex].ptr)

#define CSM_SCREEN(s)						           \
    CSMScreen *gs = GET_CSM_SCREEN (s, GET_CSM_DISPLAY (s->display))

#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))

static CompOption *
csmGetDisplayOptions (CompDisplay *display,
			 int	     *count)
{
    CSM_DISPLAY (display);

    *count = NUM_OPTIONS (gd);
    return gd->opt;
}
static gboolean csm_compare_option(CompOption * o, CompOption * p)
{
    //check if o == p
    if (o->type!=p->type)
        return FALSE;
    if (o->type==CompOptionTypeList && 
            (o->value.list.type!=p->value.list.type ||
             o->value.list.nValue!=p->value.list.nValue
             ))
        return FALSE;
    switch(o->type)
    {
        case CompOptionTypeString:
            if (!o->value.s || !p->value.s || strcmp(o->value.s,p->value.s)!=0)
                return FALSE; // if only one is NULL this isn't quite right, but it
                            //shouldn't happen
            break;
        case CompOptionTypeFloat:
            if (o->value.f!=p->value.f)
                return FALSE;
            break;
        case CompOptionTypeInt:
            if (o->value.i!=p->value.i)
                return FALSE;
            break;
        case CompOptionTypeColor:
            {
                gint i;
                for(i=0;i<4;i++)
                    if (o->value.c[i]!=p->value.c[i])
                        return FALSE;
            }
            break;
        case CompOptionTypeBool:
            if (o->value.b!=p->value.b)
                return FALSE;
            break;
        case CompOptionTypeAction:
            if (o->value.action.type!=p->value.action.type)
                return FALSE;
            if (o->value.action.button.button!=p->value.action.button.button)
                return FALSE;
            if (o->value.action.button.modifiers!=p->value.action.button.modifiers)
                return FALSE;
            if (o->value.action.key.keycode!=p->value.action.key.keycode)
                return FALSE;
            if (o->value.action.key.modifiers!=p->value.action.key.modifiers)
                return FALSE;
            if (o->value.action.edgeMask!=p->value.action.edgeMask)
                return FALSE;
            if (o->value.action.bell!=p->value.action.bell)
                return FALSE;
            break;
        case CompOptionTypeList:
            {
                gint i;
                for (i=0;i<o->value.list.nValue;i++)
                {
                    CompOption a,b;
                    a.type=o->value.list.type;
                    b.type=p->value.list.type;
                    a.value=o->value.list.value[i];
                    b.value=p->value.list.value[i];
                    if (!csm_compare_option(&a,&b))
                        return FALSE;
                }
            }
    }
    return TRUE;
}
static void
csmSetOption (CompDisplay *d,
		CompOption  *o,
		gboolean    screen,
		gchar	    *plugin)
{
    gchar      *key;
    gchar      *key2;
    CompOption p;

    CSM_DISPLAY (d);

    if (!screen)
        key = g_strconcat("a_",o->name,NULL);
    else
        key = g_strconcat("s_",o->name,NULL);
    key2 = g_strconcat("d_",key,NULL);
    if (!plugin)
        plugin="_";
    p.name=o->name;
    p.type=o->type;
    if (o->type==CompOptionTypeString)
        p.value.s=NULL;
    if (o->type==CompOptionTypeList)
    {
        p.value.list.type=o->value.list.type;
        p.value.list.nValue=0;
        p.value.list.value=0;
    }
    p.rest=o->rest;
    if ((!csm_load_option(gd->settingsFile,plugin,key,o,&p.value,d)
            && !csm_load_option(gd->schemaFile,plugin,key2,o,&p.value,d))
            || !csm_compare_option(o,&p))
    {
        /*if (screen && strcmp(screen,"allscreens")!=0)
        {
            if (plugin)
                d->screens->setScreenOptionForPlugin(d->screens,plugin,p.name,&(p.value));
            else
                d->screens->setScreenOption(d->screens,p.name,&(p.value));
        }
        else
        {
            if (plugin)
                d->setDisplayOptionForPlugin(d,plugin,p.name,&(p.value));
            else
                d->setDisplayOption(d,p.name,&(p.value));
        }*/

        csm_save_option(gd->settingsFile,plugin,key,&p,d);
        {
            gchar * fnam = g_strconcat(g_get_home_dir(),
                    "/.compiz/csm_settings",NULL);
            g_file_set_contents(fnam,
                  g_key_file_to_data(gd->settingsFile,NULL,NULL),-1,NULL);
            g_free(fnam);
        }
    }
    if (p.type==CompOptionTypeList && p.value.list.nValue && p.value.list.value)
    {
        if (p.value.list.type==CompOptionTypeString)
        {
            gint i;
            for (i=0;i<p.value.list.nValue;i++)
                free(p.value.list.value[i].s);
        }
        free(p.value.list.value);
    }
    if (p.type==CompOptionTypeString)
        free(p.value.s);
    g_free(key);
    g_free(key2);
}

static void
csmInitOption (CompDisplay *d,
		 CompOption  *o,
		 gboolean     screen,
		 gchar       *plugin)
{
    gchar      *key, *key2;
    CompScreen *s=d->screens;
    CompOptionValue *v=malloc(sizeof(CompOptionValue));
    gboolean stat;

    CSM_DISPLAY (d);
    bzero(v,sizeof(CompOptionValue));

    if (!plugin)
        plugin="_";
    if (!screen)
        key=g_strconcat("a_",o->name,NULL);
    else
        key=g_strconcat("s_",o->name,NULL);
    key2=g_strconcat("d_",key,NULL);

    stat=csm_load_option(gd->settingsFile,plugin,key,o,v,d);
    if (!stat)
        stat=csm_load_option(gd->schemaFile,plugin,key2,o,v,d);
    if (!stat)
    {
        g_warning("bailing, couln't find a val for %s in \n"
                "[%s]->%s or\n"
                "[%s]->%s",o->name,plugin,key,plugin,key2);
        return;
    }
    if (screen)
    {
        if (strcmp(plugin,"_")!=0)
        {
            s->setScreenOptionForPlugin(s,plugin,o->name,v);
        }
        else
        {
            s->setScreenOption(s,o->name,v);
        }
    }
    else
    {
        if (strcmp(plugin,"_")!=0)
        {
            d->setDisplayOptionForPlugin(d,plugin,o->name,v);
        }
        else
        {
            d->setDisplayOption(d,o->name,v);
        }
    }
    free(v);
}
static Bool
reloadProc (CompDisplay     *d,
	  CompAction      *action,
	  CompActionState state,
	  CompOption      *option,
	  int		  nOption)
{
    CSM_DISPLAY(d);
    gchar * sect, * opt;
    CompOption * o=NULL;
    CompPlugin * p;
    CompScreen * s = NULL;
    int oc = 0;
    GError * e=NULL;
    opt=g_strconcat(g_get_home_dir(),"/.compiz/csm_settings",NULL);
    g_key_file_free(gd->settingsFile);
    gd->settingsFile=g_key_file_new();
    g_key_file_load_from_file(gd->settingsFile,opt,0,&e);
    if (e)
        g_warning("couldn't load settings file!");
    g_free(opt);
    sect=getStringOptionNamed(option,nOption,"section",0);
    if (!sect)
        return TRUE;
    opt=getStringOptionNamed(option,nOption,"option",0);
    if (!opt)
        return TRUE;
    if (strlen(opt)<3)
        return TRUE;
    if (opt[0]=='s')
    {
        s = d->screens;
    }
    p=findActivePlugin(sect);
    if (p)
    {
        if (s)
        {
    		if (p->vTable->getScreenOptions)
    		    o = (*p->vTable->getScreenOptions) (s, &oc);
        }
        else
        {
    		if (p->vTable->getDisplayOptions)
    		    o = (*p->vTable->getDisplayOptions) (d, &oc);
        }
    }
    else
    {
        if (s)
            o=compGetScreenOptions(s,&oc);
        else
            o=compGetDisplayOptions(d,&oc);
    }
    o=compFindOption(o,oc,opt+2,0);
    if (o)
    {
        g_message("Reloading for [%s]->%s(%s)",sect,opt,o->name);
        csmInitOption(d,o,opt[0]=='s',sect);
        return FALSE;
    }
    else
        g_warning("Couldn't find [%s]->%s",sect,opt);
    return TRUE;
}


static Bool
csmSetDisplayOption (CompDisplay     *d,
		       char	       *name,
		       CompOptionValue *value)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, setDisplayOption);
    status = (*d->setDisplayOption) (d, name, value);
    WRAP (gd, d, setDisplayOption, csmSetDisplayOption);

    if (status)
    {
	CompOption *option;
	int	   nOption;

	option = compGetDisplayOptions (d, &nOption);
	csmSetOption (d, compFindOption (option, nOption, name, 0),
			FALSE, 0);
    }

    return status;
}

static Bool
csmSetDisplayOptionForPlugin (CompDisplay     *d,
				char	        *plugin,
				char	        *name,
				CompOptionValue *value)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, setDisplayOptionForPlugin);
    status = (*d->setDisplayOptionForPlugin) (d, plugin, name, value);
    WRAP (gd, d, setDisplayOptionForPlugin, csmSetDisplayOptionForPlugin);

    if (status)
    {
	CompPlugin *p;

	p = findActivePlugin (plugin);
	if (p && p->vTable->getDisplayOptions)
	{
	    CompOption *option;
	    int	       nOption;

	    option = (*p->vTable->getDisplayOptions) (d, &nOption);
	    csmSetOption (d, compFindOption (option, nOption, name, 0),
			    FALSE, plugin);
	}
    }

    return status;
}

static Bool
csmSetScreenOption (CompScreen      *s,
		      char	      *name,
		      CompOptionValue *value)
{
    Bool status;

    CSM_SCREEN (s);

    UNWRAP (gs, s, setScreenOption);
    status = (*s->setScreenOption) (s, name, value);
    WRAP (gs, s, setScreenOption, csmSetScreenOption);

    if (status)
    {
	CompOption *option;
	int	   nOption;

	option = compGetScreenOptions (s, &nOption);
	csmSetOption (s->display, compFindOption (option, nOption, name, 0),
			TRUE, 0);
    }

    return status;
}

static Bool
csmSetScreenOptionForPlugin (CompScreen      *s,
			       char	       *plugin,
			       char	       *name,
			       CompOptionValue *value)
{
    Bool status;

    CSM_SCREEN (s);

    UNWRAP (gs, s, setScreenOptionForPlugin);
    status = (*s->setScreenOptionForPlugin) (s, plugin, name, value);
    WRAP (gs, s, setScreenOptionForPlugin, csmSetScreenOptionForPlugin);

    if (status)
    {
	CompPlugin *p;

	p = findActivePlugin (plugin);
	if (p && p->vTable->getScreenOptions)
	{
	    CompOption *option;
	    int	       nOption;

	    option = (*p->vTable->getScreenOptions) (s, &nOption);
	    csmSetOption (s->display,
			    compFindOption (option, nOption, name, 0),
			    TRUE, plugin);
	}
    }

    return status;
}

static Bool
csmInitPluginForDisplay (CompPlugin  *p,
			   CompDisplay *d)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, initPluginForDisplay);
    status = (*d->initPluginForDisplay) (p, d);
    WRAP (gd, d, initPluginForDisplay, csmInitPluginForDisplay);

    if (status && p->vTable->getDisplayOptions)
    {
	CompOption *option;
	int	   nOption;

	option = (*p->vTable->getDisplayOptions) (d, &nOption);
	while (nOption--)
	    csmInitOption (d, option++, FALSE, p->vTable->name);
    }

    return status;
}

static Bool
csmInitPluginForScreen (CompPlugin *p,
			  CompScreen *s)
{
    Bool status;

    CSM_SCREEN (s);

    UNWRAP (gs, s, initPluginForScreen);
    status = (*s->initPluginForScreen) (p, s);
    WRAP (gs, s, initPluginForScreen, csmInitPluginForScreen);

    if (status && p->vTable->getScreenOptions)
    {
	CompOption *option;
	int	   nOption;

	option = (*p->vTable->getScreenOptions) (s, &nOption);
	while (nOption--)
	    csmInitOption (s->display, option++, TRUE, p->vTable->name);
    }

    return status;
}

static Bool
csmTimeout (void *closure)
{
    while (g_main_pending ()) g_main_iteration (FALSE);

    return TRUE;
}
static void
csmDisplayInitOptions (CSMDisplay *gd)
{
    CompOption *o;

    o = &gd->opt[CSM_DISPLAY_OPTION_RELOAD];
    o->name			  = "reload";
    o->shortDesc		  = N_("Reload");
    o->longDesc			  = N_("DBUS Signal Target for Reload");
    o->type			  = CompOptionTypeAction;
    o->value.action.initiate	  = reloadProc;
    o->value.action.terminate	  = 0;
    o->value.action.bell	  = FALSE;
    o->value.action.edgeMask	  = 0;
    o->value.action.state	  = 0;
    o->value.action.type	  = CompBindingTypeNone;
}
static void import_key_file(GKeyFile * dest, gchar * source)
{
    GKeyFile * f = g_key_file_new();
    gchar ** gr;
    gint i;
    g_key_file_load_from_file(f,source,0,NULL);
    gr = g_key_file_get_groups(f,NULL);
    for (i=0;gr && gr[i];i++)
    {
        gint j;
        gchar ** ke;
        ke = g_key_file_get_keys(f,gr[i],NULL,NULL);
        for (j=0;ke && ke[j];j++)
            g_key_file_set_value(dest,gr[i],ke[j],
                    g_key_file_get_value(f,gr[i],ke[j],NULL));
        g_strfreev(ke);
    }
    g_strfreev(gr);
    g_key_file_free(f);
}
static void
load_schema_files(CSMDisplay * gd)
{
    GDir * d;
    gchar * dpos;
    gd->schemaFile = g_key_file_new();
    import_key_file(gd->schemaFile,DATADIR "/csm/settings");
    d=g_dir_open(LIBDIR "/compiz/",0,NULL);
    if (d)
    {
        const gchar * e;
        while((e=g_dir_read_name(d)))
        {
            gchar * xfile;
            if (g_str_has_suffix(e,".csm"))
            {
                xfile = g_strconcat(g_get_home_dir(),"/.compiz/plugins/",e,NULL);
                import_key_file(gd->schemaFile,xfile);
                g_free(xfile);
            }
        }
        g_dir_close(d);
    }
    dpos = g_strconcat(g_get_home_dir(),"/.compiz/plugins/",NULL);
    d=g_dir_open(dpos,0,NULL);
    if (d)
    {
        const gchar * e;
        while((e=g_dir_read_name(d)))
        {
            gchar * xfile;
            if (g_str_has_suffix(e,".csm"))
            {
                xfile = g_strconcat(g_get_home_dir(),"/.compiz/plugins/",e,NULL);
                import_key_file(gd->schemaFile,xfile);
                g_free(xfile);
            }
        }
        g_dir_close(d);
    }
    g_free(dpos);
}
static Bool
csmInitDisplay (CompPlugin  *p,
		  CompDisplay *d)
{
    CompOption   *option;
    int	         nOption;
    CSMDisplay *gd;
    gchar * smmser;

    gd = malloc (sizeof (CSMDisplay));
    if (!gd)
	return FALSE;

    csmDisplayInitOptions (gd);

    gd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (gd->screenPrivateIndex < 0)
    {
	free (gd);
	return FALSE;
    }

    WRAP (gd, d, initPluginForDisplay, csmInitPluginForDisplay);
    WRAP (gd, d, setDisplayOption, csmSetDisplayOption);
    WRAP (gd, d, setDisplayOptionForPlugin, csmSetDisplayOptionForPlugin);

    d->privates[displayPrivateIndex].ptr = gd;

    load_schema_files(gd);
    gd->settingsFile=g_key_file_new();
    smmser=g_strconcat(g_get_home_dir(),"/.compiz",NULL);
    g_mkdir_with_parents(smmser,00755);
    g_chmod(smmser,00755);
    g_free(smmser);
    smmser=g_strconcat(g_get_home_dir(),"/.compiz/csm_settings",NULL);
    g_key_file_load_from_file(gd->settingsFile,smmser,0,NULL);
    g_free(smmser);

    option = compGetDisplayOptions (d, &nOption);
    while (nOption--)
	csmInitOption (d, option++, FALSE, 0);

    gd->timeoutHandle = compAddTimeout (KEY_CHANGE_TIMEOUT, csmTimeout, 0);

    return TRUE;
}

static void
csmFiniDisplay (CompPlugin  *p,
		  CompDisplay *d)
{
    CSM_DISPLAY (d);

    g_key_file_free(gd->schemaFile);
    g_key_file_free(gd->settingsFile);
    compRemoveTimeout (gd->timeoutHandle);

    freeScreenPrivateIndex (d, gd->screenPrivateIndex);

    free (gd);
}

static Bool
csmInitScreen (CompPlugin *p,
		 CompScreen *s)
{
    CompOption  *option;
    int	        nOption;
    CSMScreen *gs;

    CSM_DISPLAY (s->display);

    gs = malloc (sizeof (CSMScreen));
    if (!gs)
	return FALSE;

    WRAP (gs, s, initPluginForScreen, csmInitPluginForScreen);
    WRAP (gs, s, setScreenOption, csmSetScreenOption);
    WRAP (gs, s, setScreenOptionForPlugin, csmSetScreenOptionForPlugin);

    s->privates[gd->screenPrivateIndex].ptr = gs;

    option = compGetScreenOptions (s, &nOption);
    while (nOption--)
	csmInitOption (s->display, option++, TRUE, 0);

    return TRUE;
}

static void
csmFiniScreen (CompPlugin *p,
		 CompScreen *s)
{
    CSM_SCREEN (s);

    UNWRAP (gs, s, initPluginForScreen);
    UNWRAP (gs, s, setScreenOption);
    UNWRAP (gs, s, setScreenOptionForPlugin);

    free (gs);
}

static Bool
csmInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    return TRUE;
}

static void
csmFini (CompPlugin *p)
{
    if (displayPrivateIndex >= 0)
	freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginDep csmDeps[] = {
    { CompPluginRuleAfter, "dbus" },
    { CompPluginRuleBefore, "decoration" },
    { CompPluginRuleBefore, "wobbly" },
    { CompPluginRuleBefore, "fade" },
    { CompPluginRuleBefore, "cube" },
    { CompPluginRuleBefore, "scale" }
};

CompPluginVTable csmVTable = {
    "csm",
    "CSM",
    "CSM Control Backend",
    csmInit,
    csmFini,
    csmInitDisplay,
    csmFiniDisplay,
    csmInitScreen,
    csmFiniScreen,
    0, /* InitWindow */
    0, /* FiniWindow */
    csmGetDisplayOptions, /* GetDisplayOptions */ // not forever
    0, /* SetDisplayOption */
    0, /* GetScreenOptions */
    0, /* SetScreenOption */
    csmDeps,
    sizeof (csmDeps) / sizeof (csmDeps[0])
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &csmVTable;
}
